package main

import (
	natpmp "github.com/jackpal/go-nat-pmp"
	"fmt"
	"net"
	"time"
)

// Adapt the NAT-PMP protocol to the NAT interface

// TODO:
//  + Register for changes to the external address.
//  + Re-register port mapping when router reboots.
//  + A mechanism for keeping a port mapping registered.

type natPMPClient struct {
	client    *natpmp.Client
	closeChan chan bool
}

func NewNatPMP(gateway net.IP) (nat NAT) {
	return &natPMPClient{client: natpmp.NewClient(gateway)}
}

func (n *natPMPClient) GetExternalAddress() (addr net.IP, err error) {
	response, err := n.client.GetExternalAddress()
	if err != nil {
		return
	}
	ip := response.ExternalIPAddress
	addr = net.IPv4(ip[0], ip[1], ip[2], ip[3])
	return
}

func (n *natPMPClient) GetLocalAddress() (addr net.IP, err error) {
	addr = nil
	err = fmt.Errorf("Local address not implemented in Nat-PMP")
	return
}

func (n *natPMPClient) AddPortMapping(proto string, port int, description string, timeout int) (mappedExternalPort int, err error) {

	timeout2 := timeout
	if timeout == 0 {
		timeout2 = 3600
	}

	mappedExternalPort, err = n.internalAddPortMapping(proto, port, description, timeout2)

	if timeout == 0 {
		n.closeChan = make(chan bool)

		go func() {
			d := (time.Duration(timeout2) - 60) * time.Second
			tick := time.NewTicker(d)

		loop:
			for {
				select {
				case <-tick.C:
					n.internalAddPortMapping(proto, port, description, timeout2)
				case <-n.closeChan:
					break loop
				}
			}
		}()
	}
	return
}

func (n *natPMPClient) internalAddPortMapping(proto string, port int,
	description string, timeout int) (mappedExternalPort int, err error) {
	if timeout <= 0 {
		err = fmt.Errorf("timeout must not be <= 0")
		return
	}
	// Note order of port arguments is switched between our AddPortMapping and the client's AddPortMapping.
	response, err := n.client.AddPortMapping(proto, port, port, timeout)
	if err != nil {
		return
	}
	mappedExternalPort = int(response.MappedExternalPort)
	return
}

func (n *natPMPClient) DeletePortMapping(proto string, port int) (err error) {
	// Do not renew anymore
	if n.closeChan != nil {
		close(n.closeChan)
	}

	// To destroy a mapping, send an add-port with
	// an internalPort of the internal port to destroy, an external port of zero and a time of zero.
	_, err = n.client.AddPortMapping(proto, port, 0, 0)
	return
}
